# -*- coding: utf-8 -*-

"""Top-level package for BTB."""

from btb.session import BTBSession

__author__ = """MIT Data To AI Lab"""
__email__ = 'dailabmit@gmail.com'
__version__ = '0.3.8'
__all__ = ('BTBSession', )
